from django.core.validators import RegexValidator
from django.db import models


class BusinessIdentifierField(models.CharField):
    default_validators = [
        RegexValidator(
            regex=r'^[1-9a-zA-Z_.-]+$',
            message='Business ID must be alphanumeric, dots, hyphens, and underscores only.'
        )
    ]

    def __init__(self, *args, **kwargs):
        kwargs.setdefault('max_length', 255)
        super().__init__(*args, **kwargs)



from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache
from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction
from django.db.models import Prefetch

from dataviewer.models import BusinessToDataFieldMap
from metadata.services import create_change_with_reason


def get_model_class(model_name):
    cache_key = f"content_type:{model_name}"
    model_class = cache.get(cache_key)
    if not model_class:
        try:
            lower_case_model_name = model_name.lower()
            model_class = ContentType.objects.get(
                model=lower_case_model_name
            ).model_class()
            cache.set(cache_key, model_class)
        except ContentType.DoesNotExist as e:
            raise ContentType.DoesNotExist(
                f"ContentType with model {model_name} does not exist"
            ) from e
    return model_class


def get_biz_to_data_field_map(key):
    cache_key = cache_key_for_biz_to_data_field_map(key)
    field_model_map = cache.get(cache_key)
    if not field_model_map:
        try:
            data_map = BusinessToDataFieldMap.objects.get(id=key)
            field_model_map = data_map.map
            cache_biz_to_data_field_map(key, field_model_map)
        except BusinessToDataFieldMap.DoesNotExist as e:
            raise BusinessToDataFieldMap.DoesNotExist(
                f"BusinessToDataFieldMap with id {key} does not exist"
            ) from e
    return field_model_map


def create_biz_to_data_field_map(key, field_model_map, description=""):
    """
    Example:
        {
            "id": {
                "model": "WorkScope", "field": "id"
            },
            "description": {
                "model": "HistorizedWorkScopeDescription", "type": "prefetch_related",
                "field": "description",
                "order_by": "-from_epoch"
            }
        }
    """
    with transaction.atomic():
        obj, created = BusinessToDataFieldMap.objects.get_or_create(
            id=key, defaults={"map": field_model_map, "description": description}
        )
        if not created:
            raise ValueError(
                f"BusinessToDataFieldMap with id {key} already exists. Use delete_and_create_biz_to_data_field_map instead."
            )
        return obj


def delete_and_create_biz_to_data_field_map(
    key, reason_to_rewrite, field_model_map, description=""
):
    with transaction.atomic():
        # delete from database
        BusinessToDataFieldMap.objects.filter(id=key).delete()
        # create from database
        change = create_change_with_reason(reason_to_rewrite)
        # update the cache
        data_map_model = BusinessToDataFieldMap.objects.create(
            id=key, map=field_model_map, description=description, metadata=change
        )
        field_model_map = data_map_model.map
        cache_biz_to_data_field_map(key, field_model_map)


def cache_key_for_biz_to_data_field_map(key):
    return f"biz_to_data_field_map:{key}"


def cache_biz_to_data_field_map(key, model_dot_map=None):
    if model_dot_map is None:
        model_dot_map = {}
    cache_key = cache_key_for_biz_to_data_field_map(key)
    cache.set(cache_key, model_dot_map)


def get_hydrated_anchor_based_on_data_map(
    anchor_pk, main_model_class, field_model_map, fields
):
    """
    This only works when the map follows BusinessToDataFieldMap.
    Returns the hydrated instance
    """

    queryset = get_hydrated_queryset_based_on_data_map(
        {"id": anchor_pk}, main_model_class, field_model_map, fields
    )
    return queryset.first()


def get_hydrated_queryset_based_on_data_map(
    filter_parameter_and_values, main_model_class, field_model_map, fields
):
    """
    This only works when the map follows BusinessToDataFieldMap.
    Returns the hydrated instance
    """

    select_related_fields = []
    prefetch_related_fields = []
    only_fields = []

    for field in fields:
        if model_info := field_model_map.get(field):
            (only_fields, model_class) = append_only_fields_get_model_class(
                model_info, main_model_class, only_fields
            )

            (
                select_related_fields,
                prefetch_related_fields,
            ) = append_select_or_prefetch_related_fields(
                field,
                model_info,
                model_class,
                select_related_fields,
                prefetch_related_fields,
            )
            queryset = main_model_class.objects.only(*only_fields).filter(
                **filter_parameter_and_values
            )

    if select_related_fields:
        queryset = queryset.select_related(*select_related_fields)

    if prefetch_related_fields:
        queryset = queryset.prefetch_related(*prefetch_related_fields)

    return queryset


def append_only_fields_get_model_class(model_info, main_model_class, only_fields):
    model_name = model_info["model"]
    main_model_name = main_model_class.__name__

    if main_model_name == model_name:
        model_class = main_model_class

        field_name = model_info["field"]
        only_fields.append(field_name)
    else:
        model_class = get_model_class(model_name)

    return only_fields, model_class


def append_select_or_prefetch_related_fields(
    field,
    model_info,
    model_class,
    select_related_fields,
    prefetch_related_fields,
    custom_filters=[],
):
    fetch_type = model_info.get("type")
    field_name = model_info["field"]
    order_by = model_info.get("order_by", "")

    if fetch_type == "select_related":
        select_related_fields.append(field_name)
    elif fetch_type == "prefetch_related":
        related_name = model_info.get("related_name", field)
        prefetch = Prefetch(
            related_name,
            queryset=model_class.objects.only(field_name).order_by(order_by),
        )
        prefetch_related_fields.append(prefetch)

    if model_info.get("is_related_field"):
        lookup_type = model_info.get("related_lookup_type", "filter")
        if custom_filter := model_info.get("custom_filter"):
            custom_filters[custom_filter] = {
                "field": field,
                "lookup_type": lookup_type,
                "model_class": model_class,
                # other needed info
            }

    return select_related_fields, prefetch_related_fields


def transform_hydrated_instance_into_dict(anchor, field_model_map, fields):
    result = {}
    for field in fields:
        if model_info := field_model_map.get(field):
            result = append_result_with_right_data_in_field(
                anchor, model_info, field, result
            )
    return result


def transform_many_hydrated_instances(anchors, field_model_map, fields):
    return [
        transform_hydrated_instance_into_dict(anchor, field_model_map, fields)
        for anchor in anchors
    ]


def append_result_with_right_data_in_field(anchor, model_info, field, result):
    field_name = model_info["field"]
    fetch_type = model_info.get("type")

    if fetch_type == "select_related" or fetch_type != "prefetch_related":
        result[field] = getattr(anchor, field_name, None)
    else:
        related_name = model_info.get("related_name", field)
        try:
            if related_object := getattr(anchor, related_name).first():
                result[field] = getattr(related_object, field_name, None)
        except (AttributeError, ObjectDoesNotExist):
            result[field] = None
    return result



# Generated by Django 4.2.6 on 2023-10-18 11:37

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('metadata', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='DataChangeReason',
            fields=[
                ('value', models.TextField()),
                ('anchor', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='reason', serialize=False, to='metadata.datachange')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# Generated by Django 4.2.6 on 2023-10-18 09:37

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='DataChange',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='FunctionCall',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RequestCall',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RequestUrl',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('is_active', models.BooleanField(default=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ServiceFunction',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('is_active', models.BooleanField(default=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='FunctionCallArgs',
            fields=[
                ('value', models.JSONField(default=dict)),
                ('anchor', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='function_args', serialize=False, to='metadata.functioncall')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RequestCallArgs',
            fields=[
                ('value', models.JSONField(default=dict)),
                ('anchor', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='request_args', serialize=False, to='metadata.requestcall')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RequestUser',
            fields=[
                ('anchor', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='request_user', serialize=False, to='metadata.requestcall')),
                ('value', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RequestCallUrl',
            fields=[
                ('anchor', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='request_url', serialize=False, to='metadata.requestcall')),
                ('value', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='metadata.requesturl')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='FunctionCallServiceFunction',
            fields=[
                ('anchor', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='function_signature', serialize=False, to='metadata.functioncall')),
                ('value', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='metadata.servicefunction')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



from django.conf import settings
from django.db import models

from django_anchor_modeling.models import (
    AnchorNoBusinessId,
    CreatedModel,
    Knot,
    static_attribute,
)


class DataChange(AnchorNoBusinessId, CreatedModel):
    pass


StaticAttributeForDataChangeReason = static_attribute(
    anchor_class=DataChange,
    value_type=models.TextField(),
    related_name="reason",
)


class DataChangeReason(StaticAttributeForDataChangeReason):
    pass


class RequestCall(AnchorNoBusinessId):
    pass


StaticAttributeForRequestCallArgs = static_attribute(
    anchor_class=RequestCall,
    value_type=models.JSONField(default=dict),
    related_name="request_args",
)


class RequestCallArgs(StaticAttributeForRequestCallArgs):
    pass


StaticAttributeForRequestUser = static_attribute(
    anchor_class=RequestCall,
    value_type=models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE),
    related_name="request_user",
)


class RequestUser(StaticAttributeForRequestUser):
    pass


class RequestUrl(Knot):
    pass


StaticAttributeForRequestCallUrl = static_attribute(
    anchor_class=RequestCall,
    value_type=models.ForeignKey(RequestUrl, on_delete=models.CASCADE),
    related_name="request_url",
)


class RequestCallUrl(StaticAttributeForRequestCallUrl):
    pass


class FunctionCall(AnchorNoBusinessId):
    pass


StaticAttributeForFunctionCallArgs = static_attribute(
    anchor_class=FunctionCall,
    value_type=models.JSONField(default=dict),
    related_name="function_args",
)


class FunctionCallArgs(StaticAttributeForFunctionCallArgs):
    pass


class ServiceFunction(Knot):
    pass


StaticAttributeForFunctionCallSvcFunc = static_attribute(
    anchor_class=FunctionCall,
    value_type=models.ForeignKey(ServiceFunction, on_delete=models.CASCADE),
    related_name="function_signature",
)


class FunctionCallServiceFunction(StaticAttributeForFunctionCallSvcFunc):
    pass



import contextlib
from typing import Union

import pytz
from django.conf import settings
from django.core.exceptions import (
    FieldDoesNotExist,
    FieldError,
    ImproperlyConfigured,
    ValidationError,
)
from django.db import IntegrityError, models, transaction
from django.db.models.constants import LOOKUP_SEP
from django.db.models.utils import resolve_callables
from django.utils import timezone


# Higher-order function to generate manager methods
def generate_get_by_parent_method(related_name: str):
    def get_by_parent(self, parent_instance: Union[models.Model, int]):
        filter_args_and_values = {f"{related_name}__value": parent_instance}
        if isinstance(parent_instance, int):
            filter_args_and_values = {f"{related_name}__value_id": parent_instance}
        return self.filter(**filter_args_and_values)

    return get_by_parent


def add_method_to_manager(manager, related_name):
    def get_by_related_name(self, parent_instance: Union[models.Model, int]):
        filter_args_and_values = {f"{related_name}__value": parent_instance}
        if isinstance(parent_instance, int):
            filter_args_and_values = {f"{related_name}__value_id": parent_instance}
        return self.filter(**filter_args_and_values)

    setattr(manager, f"get_by_{related_name}", get_by_related_name)


def create_prepare_filter_manager(related_name):
    class PrepareFilterManager(models.Manager):
        def get_by_related_name(self, parent_instance: Union[models.Model, int]):
            filter_args_and_values = {f"{related_name}__value": parent_instance}
            if isinstance(parent_instance, int):
                filter_args_and_values = {f"{related_name}__value_id": parent_instance}
            return self.filter(**filter_args_and_values)

    return PrepareFilterManager()


# class BusinessEventFiltersManager(models.Manager):
#     """
#     This is to generate dynamic filter_args_and_values for BusinessEvent
#     such as {related_name_of_parent__value: parent_instance}
#     or {related_name_of_parent__value_id: parent_instance.pk}
#     """

#     pass
#
#
# After the parent is craeted by way of static_attribute
# then we do
# # as parent filtering the child by parent is a common query
# Attach the dynamically generated method to the manager
# BusinessEventFiltersManager.get_by_work_scope = generate_get_by_parent_method(
#     "parent_work_scope"
# )


class PrepareFilterManager(models.Manager):
    pass


class ZeroUpdateStrategyManager(models.Manager):
    """
    A manager class that provides a zero update strategy for model instances.
    """

    def delete_and_create(self, pk=None, anchor=None, **kwargs):
        """
        Deletes the given instance from the database and creates a new instance with the
        provided keyword arguments.

        Args:
            pk: The primary key of the instance to delete. Used for regular models.
            anchor: The related object for OneToOneField primary key.
                    This serves as the unique identifier. Used for models
                    with OneToOneField as primary key.
            **kwargs: The keyword arguments used to create the new instance.

        Returns:
            Model: The newly created instance.
        """

        # Validate the input based on whether the model uses OneToOneField as a PK
        if isinstance(self.model._meta.pk, models.OneToOneField):
            if anchor is None:
                error_text = (
                    "anchor must be provided for models "
                    "with OneToOneField as primary key."
                )
                raise ValueError(error_text)
        elif pk is None:
            raise ValueError("pk must be provided for regular models.")

        with transaction.atomic():
            # Delete the existing instance
            if isinstance(self.model._meta.pk, models.OneToOneField):
                self.filter(pk=anchor.pk).delete()
            else:
                self.filter(pk=pk).delete()

            # Prepare the kwargs for creating a new instance
            if isinstance(self.model._meta.pk, models.OneToOneField):
                kwargs[self.model._meta.pk.name] = anchor
            else:
                kwargs["pk"] = pk

            new_instance = self.model(**kwargs)
            new_instance.avoid_recursion = True
            new_instance.save()

        return new_instance

    def create(self, **kwargs):
        """
        Creates a new instance of the model with the given keyword arguments,
        sets the 'created_via_manager' and 'avoid_recursion' attributes to True,
        saves the instance to the database, and returns the instance.

        Args:
            **kwargs: The keyword arguments used to create the instance.

        Returns:
            Model: The created instance.
        """

        instance = self.model(**kwargs)
        instance.created_via_manager = True
        instance.avoid_recursion = True  # prevent recursion in save
        instance.save()
        instance.created_via_manager = False
        instance.avoid_recursion = False  # prevent recursion in save
        return instance

    def _extract_model_params(self, defaults, **kwargs):
        """
        Replicated from Django 4.2.6 at
        https://github.com/django/django/blob/c22017bd1dddb5b57d8a179e6378ef0c8d7f5eeb/django/db/models/query.py#L981
        Prepare `params` for creating a model instance based on the given
        kwargs; for use by get_or_create().
        """
        defaults = defaults or {}
        params = {k: v for k, v in kwargs.items() if LOOKUP_SEP not in k}
        params |= defaults
        property_names = self.model._meta._property_names
        invalid_params = []
        for param in params:
            try:
                self.model._meta.get_field(param)
            except FieldDoesNotExist:
                # It's okay to use a model's property if it has a setter.
                if not (param in property_names and getattr(self.model, param).fset):
                    invalid_params.append(param)
        if invalid_params:
            raise FieldError(
                "Invalid field name(s) for model %s: '%s'."
                % (
                    self.model._meta.object_name,
                    "', '".join(sorted(invalid_params)),
                )
            )
        return params

    def get_or_create(self, defaults=None, **kwargs):
        """
        Replicated from Django 4.2.6 at
        https://github.com/django/django/blob/c22017bd1dddb5b57d8a179e6378ef0c8d7f5eeb/django/db/models/query.py#L906
        Look up an object with the given kwargs, creating one if necessary.
        Return a tuple of (object, created), where created is a boolean
        specifying whether an object was created.
        """
        try:
            return self.get(**kwargs), False
        except self.model.DoesNotExist:
            params = self._extract_model_params(defaults, **kwargs)
            # Try to create an object using passed params.
            try:
                with transaction.atomic(using=self.db):
                    params = dict(resolve_callables(params))
                    return self.create(**params), True
            except IntegrityError:
                with contextlib.suppress(self.model.DoesNotExist):
                    return self.get(**kwargs), False
                raise

    def update_or_create(self, *args, **kwargs):
        """
        Raises a NotImplementedError with a message indicating that update_or_create is not
        allowed in ZeroUpdateStrategyManager.

        Raises:
            NotImplementedError: Raised when update_or_create is called in
            ZeroUpdateStrategyManager.

        Example:
            ```python
            manager = ZeroUpdateStrategyManager()
            try:
                manager.update_or_create()
            except NotImplementedError as e:
                print(str(e))  # Output: update_or_create is not allowed in ZeroUpdateStrategyManager
            ```
        """

        raise NotImplementedError(
            "update_or_create is not allowed in ZeroUpdateStrategyManager"
        )

    def bulk_create(self, *args, **kwargs):
        """
        Raises a NotImplementedError with a message indicating that bulk_create is not
        allowed in ZeroUpdateStrategyManager.

        Raises:
            NotImplementedError: Raised when bulk_create is called in
            ZeroUpdateStrategyManager.

        Example:
            ```python
            manager = ZeroUpdateStrategyManager()
            try:
                manager.bulk_create()
            except NotImplementedError as e:
                print(str(e))  # Output: bulk_create is not allowed in ZeroUpdateStrategyManager
            ```
        """

        raise NotImplementedError(
            "bulk_create is not allowed in ZeroUpdateStrategyManager"
        )

    def bulk_update(self, *args, **kwargs):
        """
        Raises a NotImplementedError with a message indicating that bulk_update is not
        allowed in ZeroUpdateStrategyManager.

        Raises:
            NotImplementedError: Raised when bulk_update is called in
            ZeroUpdateStrategyManager.

        Example:
            ```python
            manager = ZeroUpdateStrategyManager()
            try:
                manager.bulk_update()
            except NotImplementedError as e:
                print(str(e))  # Output: bulk_update is not allowed in ZeroUpdateStrategyManager
            ```
        """

        raise NotImplementedError(
            "bulk_update is not allowed in ZeroUpdateStrategyManager"
        )

    def update(self, *args, **kwargs):
        """
        Raises a NotImplementedError with a message indicating that update is not
        allowed in ZeroUpdateStrategyManager.

        Raises:
            NotImplementedError: Raised when update is called in
            ZeroUpdateStrategyManager.

        Example:
            ```python
            manager = ZeroUpdateStrategyManager()
            try:
                manager.update()
            except NotImplementedError as e:
                # Output: update is not allowed in ZeroUpdateStrategyManager
                print(str(e))
            ```
        """
        raise NotImplementedError("update is not allowed in ZeroUpdateStrategyManager")


class CompositeKeyManager(ZeroUpdateStrategyManager):
    """
    A manager class for models that use a composite key.

    Attributes:
        check_for_from_epoch (bool): Indicates whether to check for the presence of
        'from_epoch' in the composite key fields.
        Defaults to False.

    """

    check_for_from_epoch = False

    def create_composite_key(self, instance):
        """
        Creates a composite key for given instance based on
        composite_key_fields option defined in its model class.

        The order follows the order in which the fields are defined in
        composite_key_fields option.

        Args:
            self: The manager instance.
            instance: The instance for which to create the composite key.

        Returns:
            str: Composite key generated from the foreign key values of the instance.

        """
        composite_key_fields_option = instance.composite_key_fields

        if not composite_key_fields_option:
            raise ImproperlyConfigured(
                f"{instance._meta.model_name} must define composite_key_fields."
            )

        # Handle the different ways composite_key_fields can be defined
        if isinstance(composite_key_fields_option, (list, tuple)):
            composite_key_fields_fields = composite_key_fields_option
        else:
            composite_key_fields_fields = []

        if not composite_key_fields_fields:
            raise ImproperlyConfigured(
                f"composite_key_fields in {instance._meta.model_name} cannot be empty"
            )

        if (
            self.check_for_from_epoch
            and "from_epoch" not in composite_key_fields_fields
        ):
            error_message = (
                f"composite_key_fields in {instance._meta.model_name} must contain"
                " 'from_epoch'"
                " because {instance._meta.model_name} is a historized model."
            )
            raise ImproperlyConfigured(error_message)

        foreign_keys = [
            field
            for field in composite_key_fields_fields
            if isinstance(instance._meta.get_field(field), models.ForeignKey)
        ]

        composite_key_parts = []
        for fk in foreign_keys:
            fk_value = getattr(instance, f"{fk}_id", None)
            if fk_value is not None:
                composite_key_parts.append(str(fk_value))

        return ".".join(composite_key_parts)

    def create(self, **kwargs):
        """
        Creates a new instance of the model with the given keyword arguments, assigns a composite key to the instance, saves it to the database, and returns the instance.

        Args:
            **kwargs: The keyword arguments used to create the instance.

        Returns:
            Model: The created instance.

        Example:
            ```python
            manager = MyModelManager()
            instance = manager.create(field1='value1', field2='value2')
            print(instance.id)  # Output: <composite key value>
            ```
        """
        try:
            with transaction.atomic():
                instance = self.model(**kwargs)
                instance.id = self.create_composite_key(instance)
                return super().create(**kwargs)
        except IntegrityError as e:
            if "unique constraint" in str(e).lower() and "id" in str(e).lower():
                raise ValidationError(
                    f"A record with the composite key {instance.id} already exists."
                ) from e
            raise  # Re-raise the original exception if it wasn't due to id uniqueness


class FromModelManager(ZeroUpdateStrategyManager):
    """
    A manager class for any model that inherits FromModel
    """

    def timezone_create(self, tz, **kwargs):
        """
        Creates a new instance with the given timezone, anchor, and value.

        Args:
            tz: The timezone for the instance.
            anchor: The anchor for the instance.
            value: The value for the instance.
            **kwargs: Additional attributes for the instance.

        Returns:
            The newly created instance.
        """
        return self.create_with_timezone(pytz.timezone(tz), **kwargs)

    def create_with_timezone(self, tz, **kwargs):
        """
        Creates a new instance with the given timezone, and additional attributes.

        Args:
            tz: The timezone for the instance.
            **kwargs: Additional attributes for the instance.

        Returns:
            The newly created instance.
        """
        kwargs |= {
            "from_epoch": int(timezone.now().timestamp()),
            "from_utc_start": timezone.now().strftime("%Y-%m-%dT%H:%M:%S%z"),
            "from_local_start": timezone.now()
            .astimezone(tz)
            .strftime("%Y-%m-%dT%H:%M:%S"),
            "from_timezone_id": str(tz),
            "from_timezone_rules": pytz.OLSON_VERSION,
        }
        return super().create(**kwargs)

    def create(self, **kwargs):
        """
        Creates a new instance with the given additional attributes.
        The instance is created with the default timezone.

        Args:
            **kwargs: Additional attributes for the instance.

        Returns:
            The newly created instance.
        """
        default_tz = pytz.timezone(settings.TIME_ZONE)
        return self.create_with_timezone(default_tz, **kwargs)


class HistorizedAttributeManager(ZeroUpdateStrategyManager):
    """
    A manager class for HistorizedAttribute model
    """

    def timezone_create(self, tz, anchor, value, **kwargs):
        """
        Creates a new instance with the given timezone, anchor, and value.

        Args:
            tz: The timezone for the instance.
            anchor: The anchor for the instance.
            value: The value for the instance.
            **kwargs: Additional attributes for the instance.

        Returns:
            The newly created instance.
        """
        return self.create_with_timezone_and_anchor(
            pytz.timezone(tz), anchor, value, **kwargs
        )

    def create_with_timezone_and_anchor(self, tz, anchor, value, **kwargs):
        """
        Creates a new instance with the given timezone, anchor, value.
        Note that the primary key is a composite of anchor and timestamp.

        Args:
            tz: The timezone for the instance.
            anchor: The anchor for the instance.
            value: The value for the instance.
            **kwargs: Additional attributes for the instance.

        Returns:
            The newly created instance.
        """
        current_time = timezone.now()
        from_epoch = int(current_time.timestamp())
        local_time = current_time.astimezone(tz)

        kwargs |= {
            "id": f"{str(anchor.id)}.{from_epoch}",
            "anchor": anchor,
            "value": value,
            "from_epoch": from_epoch,
            "from_utc_start": current_time.strftime("%Y-%m-%dT%H:%M:%S%z"),
            "from_local_start": local_time.strftime("%Y-%m-%dT%H:%M:%S%z"),
            "from_timezone_id": str(tz),
            "from_timezone_rules": pytz.OLSON_VERSION,
        }
        return super().create(**kwargs)

    def create(self, anchor, value, **kwargs):
        """
        Creates a new instance with the given anchor, value, and additional attributes.
        The instance is created with the default timezone.

        Args:
            anchor: The anchor for the instance.
            value: The value for the instance.
            **kwargs: Additional attributes for the instance.

        Returns:
            The newly created instance.
        """
        default_tz = pytz.timezone(settings.TIME_ZONE)
        return self.create_with_timezone_and_anchor(default_tz, anchor, value, **kwargs)


class HistorizedTieManager(FromModelManager, CompositeKeyManager):
    """
    A manager class for the HistorizedTie model that combines the functionality of
    FromModelManager and CompositeKeyManager.

    """

    check_for_from_epoch = True



from types import new_class

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.validators import RegexValidator
from django.db import models
from django.db.utils import IntegrityError

from .fields import BusinessIdentifierField
from .managers import (
    CompositeKeyManager,
    FromModelManager,
    HistorizedAttributeManager,
    PrepareFilterManager,
    ZeroUpdateStrategyManager,
    add_method_to_manager,
    create_prepare_filter_manager,
)


class TimeStampedModel(models.Model):
    """
    An abstract base class model that provides selfupdating
    ``created`` and ``modified`` fields.
    """

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class PersonStampedModel(models.Model):
    """
    An abstract base class model that provides self-updating
    ``created_by`` and ``modified_by`` fields.
    """

    created_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        editable=False,
        related_name="created_%(app_label)s_%(class)s",
        related_query_name="query_created_%(app_label)s_%(class)ss",
        on_delete=models.CASCADE,
    )
    modified_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        null=True,
        editable=False,
        related_name="last_edited_%(app_label)s_%(class)s",
        related_query_name="query_last_edited_%(app_label)s_%(class)ss",
        on_delete=models.CASCADE,
    )

    class Meta:
        abstract = True

    @property
    def activity_actor_attr(self):
        return self.modified_by or self.created_by


class CreatedModel(models.Model):
    """
    An abstract base class model that provides selfupdating
    ``created`` field ONLY
    """

    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        abstract = True


class UndeletableModel(models.Model):
    """
    An abstract base class model that disallows delete
    probably needs another that disallows delete depending on conditions
    """

    def delete(self):
        pass

    class Meta:
        abstract = True


class ZeroUpdateStrategyModel(models.Model):
    """
    An abstract base class model that provides a zero update strategy for saving instances.

    Attributes:
        objects (ZeroUpdateStrategyManager): The manager instance for the model.

    Meta:
        abstract (bool): Specifies that this model is an abstract base class.

    Methods:
        save: Overrides the save method to apply the zero update strategy.

    Args:
        self: The model instance.
        *args: Additional positional arguments.
        **kwargs: Additional keyword arguments.
    """

    objects = ZeroUpdateStrategyManager()
    filters = PrepareFilterManager()

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        """
        Saves the model instance to the database.

        If 'avoid_recursion' attribute is present, it calls the superclass's save method
        and removes the attribute.

        If the instance already has a primary key and the 'created_via_manager'
        attribute is not present, it deletes the existing instance and creates a new one
        with the same field values.

        Otherwise, it calls the superclass's save method.

        Args:
            *args: Additional positional arguments to pass to the superclass's save method.
            **kwargs: Additional keyword arguments to pass to the superclass's save method.

        Returns:
            None
        """
        if hasattr(self, "avoid_recursion") and self.avoid_recursion:
            super().save(*args, **kwargs)
            delattr(self, "avoid_recursion")
        elif self.pk is not None and (
            not hasattr(self, "created_via_manager") or not self.created_via_manager
        ):
            field_names = [field.name for field in self._meta.get_fields()]
            filtered_dict = {k: v for k, v in self.__dict__.items() if k in field_names}
            type(self).objects.delete_and_create(pk=self.pk, **filtered_dict)
        else:
            super().save(*args, **kwargs)


class CharFieldForCompositeKey(models.Model):
    """
    An abstract base class for a CharField used as a composite key in Django models.

    Attributes:
        id (CharField): The primary key field.

    Meta:
        abstract (bool): Indicates that this model is abstract and cannot be instantiated.
        composite_key_fields (tuple): The fields that make up the composite key.

    Example:
        ```python
        class MyModel(CharFieldForCompositeKey):

            field1 = models.CharField(max_length=255)
            field2 = models.CharField(max_length=255)

            composite_key_fields = ('field1', 'field2')
        ```
    """

    composite_key_fields = ()
    id = models.CharField(
        max_length=255,
        primary_key=True,
        validators=[
            RegexValidator(
                regex=r"^[0-9a-zA-Z_.-]+$",
                message="ID must be alphanumeric, underscores, hyphens, and dots.",
            )
        ],
    )

    class Meta:
        abstract = True


class FromModel(ZeroUpdateStrategyModel):
    """
    An abstract base class model that provides fields for storing datetime information.

    It sets datetime fields based on Jon Skeet's advice on datetime storage
    at https://codeblog.jonskeet
    .uk/2019/03/27/storing-utc-is-not-a-silver-bullet/
    while conforming to anchor modeling's use of from_date.

    Works for anchor modeling's `attribute`, `tie`, and `knotted` models.

    This FromModel alone will not make a model "historized".

    Please see the documentation for `HistorizedAttribute` for more information.

    Attributes:
        from_epoch (IntegerField): An IntegerField that stores the epoch time.
        from_utc_start (DateTimeField): A DateTimeField that stores the UTC start time.
        from_local_start (DateTimeField): A DateTimeField that stores local start time.
        from_timezone_id (CharField): A CharField that stores the timezone ID.
        from_timezone_rules (CharField): A CharField that stores the timezone rules.

        disallowed_exception (Exception): Exception raised when create record via save.

    Example:
        from_epoch = 1625904000
        from_utc_start = "2021-07-10T08:00:00Z"
        from_local_start = "2021-07-10T09:00:00"
        from_timezone_id = "Europe/Amsterdam"
        from_timezone_rules = "2020c"


    Meta:
        abstract (bool): Specifies that this model is an abstract base class.

    Usage:
        ```
        class MyConcreteModel(FromModel, OtherAbstractModel):
            objects = OtherModelManager()
            from_objects = FromModelManager() # strongly recommended

            class Meta:
                abstract = False
        ```

        Do NOT use `save` directly for creating new records.
        You will trigger `FromCannotCreateRecordViaSave` exception.

        Using `save` to update is fine.

        Use any of the following instead:
         - `MyConcreteModel.from_objects.create`
         - `MyConcreteModel.from_objects.create_with_timezone`
         - `MyConcreteModel.from_objects.timezone_create`
         - `MyConcreteModel.from_objects.user_create`
    """

    from_epoch = models.IntegerField(editable=False)
    from_utc_start = models.DateTimeField(editable=False)
    from_local_start = models.DateTimeField(editable=False)
    from_timezone_id = models.CharField(max_length=50, editable=False)
    from_timezone_rules = models.CharField(max_length=20, editable=False)

    from_objects = FromModelManager()  # strongly recommended

    class Meta:
        abstract = True


def historized_attribute(anchor_class, value_type, related_name="%(class)s_related"):
    """
    A factory function that generates an abstract base class model for historized attributes.

    Args:
        anchor_class: The class to which the historized attribute is anchored.
        value_type: The type of the historized attribute value.
        related_name (str, optional): The related name for the anchor field. Defaults to "%(class)s_related".

    Returns:
        HistorizedAttribute: An abstract base class model for historized attributes.

    """

    class HistorizedAttribute(CharFieldForCompositeKey, FromModel):
        """
        An abstract base class model for historized attributes.

        Attributes:
            id (CharField): A CharField that represents the ID of the historized attribute.
            anchor (ForeignKey): A ForeignKey to the anchor class.
            value (value_type): A field that represents the value of the historized attribute.
            objects (HistorizedAttributeManager): The manager instance for the model.

        Meta:
            abstract (bool): Specifies that this model is an abstract base class.

        """

        anchor = models.ForeignKey(
            anchor_class, on_delete=models.CASCADE, related_name=related_name
        )
        value = value_type

        objects = HistorizedAttributeManager()

        class Meta:
            abstract = True

    return HistorizedAttribute


def static_attribute(anchor_class, value_type, related_name="%(class)s_related"):
    """
    A factory function that generates a model for static attributes.

    Args:
        anchor_class: The class to which the static attribute is anchored.
        value_type: The type of the static attribute value.
        related_name (str, optional): The related name for the anchor field. Defaults to "%(class)s_related".

    Returns:
        StaticAttribute: A model for static attributes.

    """

    class StaticAttribute(ZeroUpdateStrategyModel):
        """
        A model for static attributes.

        This also works for ForeignKey as attributes

        Attributes:
            anchor (OneToOneField): A OneToOneField to the anchor class as primary key.
            value (value_type): A field representing the value of the static attribute.
            objects (ZeroUpdateStrategyManager): The manager instance for the model.

        Meta:
            abstract (bool): Specifies that this model is an abstract base class.

        """

        anchor = models.OneToOneField(
            anchor_class,
            on_delete=models.CASCADE,
            related_name=related_name,
            primary_key=True,
        )
        value = value_type

        objects = ZeroUpdateStrategyManager()

        class Meta:
            abstract = True

    # This block is inside static_attribute but outside StaticAttribute
    if isinstance(value_type, models.ForeignKey):
        # Dynamically set the manager method on the anchor class
        # so can more easily get_by_parent_related_name
        if hasattr(anchor_class, "filters"):
            add_method_to_manager(anchor_class.filters, related_name)
        else:
            # create_custom_manager should be modified to accommodate this logic
            custom_manager = create_prepare_filter_manager(related_name)
            anchor_class.add_to_class("filters", custom_manager)

    return StaticAttribute


def is_a(super_class_anchor, business_identifier_on=True):
    """
    A factory function that generates a subtype model under an anchor
    which is a super class

    Args:
        super_class_anchor: The class to which this model is subtype to.
        business_identifier_on(boolean): If true, add business_identifier

    Returns:
        IsA: A model for subtype.

    """
    # Retrieve original manager from super_class_anchor
    original_manager = getattr(super_class_anchor, "objects", models.Manager)

    # Dynamically create a composite manager that combines the two managers
    CompositeManager = new_class(
        "CompositeManager", (original_manager.__class__, ZeroUpdateStrategyManager)
    )

    class IsA(ZeroUpdateStrategyModel):
        """
        A model for subtype. Which is like a StaticTie

        Attributes:
            super_class (OneToOneField): A OneToOneField to the anchor class as primary key.
            business_identifier (CharField): A field representing the record
            objects (ZeroUpdateStrategyManager): The manager instance for the model.

        Meta:
            abstract (bool): Specifies that this model is an abstract base class.

        """

        super_class = models.OneToOneField(
            super_class_anchor,
            on_delete=models.CASCADE,
            related_name="%(class)s",
            primary_key=True,
        )
        if business_identifier_on:
            business_identifier = BusinessIdentifierField(unique=True)

        objects = CompositeManager()

        class Meta:
            abstract = True

    return IsA


class StaticTie(CharFieldForCompositeKey, ZeroUpdateStrategyModel):
    objects = CompositeKeyManager()

    class Meta:
        abstract = True


class HistorizedTie(CharFieldForCompositeKey, FromModel):
    """
    An abstract base class for a historized tie model that uses a composite key
    and supports historization.

    Attributes:
        objects (CompositeKeyManager): The manager for the model.

    Meta:
        abstract (bool): Indicates that this model is abstract
            and cannot be instantiated.

    Example:
        ```python
        class MyHistorizedTie(HistorizedTie):

            field1 = models.CharField(max_length=255)
            field2 = models.CharField(max_length=255)

            composite_key_fields = ('field1', 'field2', 'from_epoch')
        ```
    """

    objects = CompositeKeyManager()

    class Meta:
        abstract = True


class AnchorWithBusinessId(ZeroUpdateStrategyModel):
    business_identifier = BusinessIdentifierField(unique=True)

    class Meta:
        abstract = True


class AnchorNoBusinessId(ZeroUpdateStrategyModel):
    class Meta:
        abstract = True


class KnotManager(models.Manager):
    def ensure_valid_keys_exist(self):
        model_cls = self.model
        model_name = model_cls.__name__
        valid_keys = getattr(model_cls, "VALID_KEYS", None)

        if not valid_keys:
            raise ImproperlyConfigured(
                f"'VALID_KEYS' must be defined and non-empty for {model_name}"
            )

        for key in valid_keys:
            try:
                _, created = model_cls.objects.get_or_create(pk=key)
                if created:
                    print(f"Created {model_name} with pk {key}")
                else:
                    print(f"{model_name} with pk {key} already exists")
            except IntegrityError:
                print(f"Could not create {model_name} with pk {key}. IntegrityError.")
                # re-raise the exception
                raise


class Knot(UndeletableModel):
    id = models.CharField(max_length=255, primary_key=True)
    is_active = models.BooleanField(default=True)

    objects = KnotManager()

    class Meta:
        abstract = True
